# Copyright 2021 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Iterable, TYPE_CHECKING
from numbers import Number

import numpy as np


if TYPE_CHECKING:
    import matplotlib

from ..dispatch import dispatch
from ..numbers import is_scalar
from ..types import Array, DType
from ..optional_deps import import_optional_dependency


def raise_if_len_not_match(length, expected_length, string):
    if length != expected_length:
        raise ValueError(
            f"""
            Length mismatch: expected object of length {expected_length}, but
            got object of length {length} for key {string}.
            """
        )


def maybecopy(maybe_arr):
    if isinstance(maybe_arr, np.ndarray):
        return maybe_arr.copy()
    else:
        return maybe_arr


def replace_none_with_nan(item):
    # recursively replace None with np.nan in lists
    if isinstance(item, list):
        return [replace_none_with_nan(subitem) for subitem in item]
    return np.nan if item is None else item


class History:
    """
    A class to store a time-series of arbitrary data.

    An History object stores several time-series all sharing the same
    time axis. History behaves like a dictionary, where the various key
    index the values (y axis) of the time-series. The time-axis is accessed
    through the attribute `History.iters`.

    It's possible to label one time-serie (one key) as the main value, so that
    when converting to numpy array (for example for plotting with pyplot) that
    axis is automatically picked by default.

    If only one time-series is provided, without a key, then its name will
    be `value`.
    """

    __slots__ = ("_value_dict", "_value_name", "_single_value", "_keys")

    def __init__(
        self,
        values: Any = None,
        iters: list | Array | None = None,
        dtype: DType | None = None,
        iter_dtype: DType | None = None,
        main_value_name: str | None = None,
    ):
        """
        Creates a new History object.

        Values should be an arbitrary type or container to initialize the
        History with.
        By default assumes that values correspond to the first iteration `0`.
        If `values` is a list or collection of lists, an array or range should
        be passed to values with the correct length.

        Optionally it's possible to specify the dtype of the data and of the
        time-axis.

        Args:
            values: a type/container of types containing the value at the first
                iteration, or an iterable/container of iterables
                containing the values at all iterations (in the latter, values
                must also be specified). It may also be the dictionary obtained
                from another History object, when converted to dictionary.
            iters: an optional iterable of iterations at which values correspond.
                If unspecified, assumes that values are logged at only one iteration.
            dtype: If no values or iters are passed, uses this dtype to store data
                if numerical
            iter_dtype: If no values or iters are passed, uses this dtype to store
                iteration numbers
            main_value_name: If data is a dict or object with to_dict method, this
                optional string labels an entry as being the main one.
        """
        single_value = False

        if values is None and iters is None:
            values = []
            iters = []
        elif iters is None:
            # If it's a dictionary, copy it
            if isinstance(values, dict):
                values = {k: maybecopy(v) for k, v in values.items()}
                if "iters" in values:
                    iters = values["iters"]
                    del values["iters"]
            if iters is None:
                iters = 0

        if is_scalar(iters):
            iters = np.array([iters], dtype=iter_dtype)
        elif isinstance(iters, list):
            iters = [np.nan if x is None else x for x in iters]
            iters = np.array(iters, dtype=iter_dtype)

        n_elements = len(iters)

        if is_scalar(values):
            values = {"value": values}
            single_value = True

        elif hasattr(values, "__array__"):
            values = {"value": values}
            single_value = True

        elif hasattr(values, "to_compound"):
            main_value_name, values = values.to_compound()
            single_value = len(values.keys()) == 1

        elif hasattr(values, "to_dict"):
            values = values.to_dict()
            single_value = len(values.keys()) == 1

        elif isinstance(values, dict) or hasattr(values, "items"):
            single_value = len(values.keys()) == 1

        else:
            values = {"value": [values]}
            single_value = True

        if single_value:
            main_value_name = list(values.keys())[0]

        value_dict = {"iters": iters}
        keys = []
        for key, val in values.items():
            if key == "iters":
                raise ValueError("cannot have a field called iters")

            if is_scalar(val):
                raise_if_len_not_match(1, n_elements, key)
                val = np.asarray([val], dtype=dtype)

            elif hasattr(val, "__array__"):
                val = np.asarray(val, dtype=dtype)
                if n_elements == 1 and val.ndim > 0:
                    val = np.reshape(val, (1, *val.shape))

                raise_if_len_not_match(len(val), n_elements, key)

            elif isinstance(val, list):
                # see below
                # val = np.asarray(val, dtype=dtype)

                # Recursively replace None with np.nan in nested lists
                # only relevant for loading from json on disk
                val = np.asarray(replace_none_with_nan(val), dtype=dtype)
            else:
                val = [val]

            value_dict[key] = val
            keys.append(key)

        self._value_dict = value_dict
        self._value_name = main_value_name
        self._single_value = single_value
        self._keys = keys

    @property
    def main_value_name(self) -> str | None:
        """The name of the main value in this history object, if defined."""
        return self._value_name

    @property
    def iters(self) -> Array:
        return self._value_dict["iters"]

    @property
    def values(self) -> Array:
        if self._value_name is None:
            raise ValueError("No main value defined for this history object.")
        return self._value_dict[self._value_name]

    def __len__(self) -> int:
        return len(self.iters)

    def __getitem__(self, key) -> Array:
        # if its an int corresponding to an element not inside the dict,
        # treat it as accessing a slice of a single element
        if isinstance(key, int) and key not in self:
            return self._get_slice(key)

        # support slice syntax
        if isinstance(key, slice):
            return self._get_slice(key)

        return self._value_dict[key]

    def _get_slice(self, slce: slice) -> "History":
        """
        get a slice of iterations from this history object
        """
        values_sliced = {}
        for key in self.keys():
            values_sliced[key] = self[key][slce]

        iters = self.iters[slce]

        hist = History(values_sliced, iters)
        hist._single_value = self._single_value
        hist._value_name = self._value_name

        return hist

    def __contains__(self, key: str) -> bool:
        return key in self._value_dict

    def keys(self) -> list:
        return self._keys

    def append(self, val: Any, it: Number | None = None):
        """
        Append another value to this history object.

        Args:
            val: the value in the next timestep
            it: the time corresponding to this new value. If
                not defined, increment by 1.
        """
        append(self, val, it)

    def get(self) -> tuple[Array, Array]:
        """
        Returns a tuple containing times and values of this history object
        """
        return self.iters, self.values

    def to_dict(self) -> dict:
        """
        Converts the history object to dict.

        Used for serialization
        """
        return self._value_dict

    def __array__(self, *args, **kwargs) -> Array:
        """
        Automatically transform this object to a numpy array when calling
        asarray, by only considering the values and neglecting the times.
        """
        return np.asarray(self.values, *args, **kwargs)

    def __iter__(self):
        """
        You can iterate the values in history object.

        Returns the Iterator object.
        """
        return iter(zip(self.iters, self.values))

    def __getattr__(self, attr):
        # Allow users to access fields with . accessor patterns
        if attr in self._value_dict:
            return self._value_dict[attr]

        raise AttributeError

    def __repr__(self):
        if len(self.iters) < 5:
            iters_repr = repr(self.iters)
        else:
            iters_repr = (
                f"[{self.iters[0]}, {self.iters[1]}, ..."
                f" {self.iters[-2]}, {self.iters[-1]}] "
                f"({len(self.iters)} steps)"
            )
        return (
            "History("
            + f"\n   keys  = {self.keys()}, "
            + f"\n   iters = {iters_repr},"
            + "\n)"
        )

    def __str__(self):
        return f"History(keys={self.keys()}, n_iters={len(self.iters)})"

    def plot(
        self,
        *args,
        fig: "matplotlib.figure.Figure | None" = None,
        show: bool = True,
        all: bool | None = None,
        xscale: str | None = None,
        yscale: str | None = "auto",
        **kwargs,
    ) -> "matplotlib.axes.Axes" | Iterable["matplotlib.axes.Axes"]:
        """
        Plot the history object using matplotlib.

        This function is equivalent to calling `matplotlib.pyplot.plot` on the
        values of the history object. If multiple keys are present, it will
        plot only the main value by default. If `all=True`, it will plot all
        the keys in the history object.

        When plotting a single key, this function is equivalent to calling
        :code:`plt.plot(history.iters, history[key], *args, **kwargs)`.

        For the arguments and keyword arguments, refer to the documentation of
        :func:`matplotlib.pyplot.plot`. Below we document some extra arguments.

        Args:
            fig: A matplotlib figure object to plot the data on. If not provided,
                a new figure is created.
            all: If True, plot all the keys in the history object. If False, plot
                only the main value. If None, plot only the main value if it is
                defined, otherwise plot all the keys.
            show: If True, call `plt.show()` at the end of the function. If False,
                the plot is not shown.

        """
        plt = import_optional_dependency("matplotlib.pyplot", descr="plot")

        if self.main_value_name is None and all is None:
            all = True

        if fig is None:
            fig = plt.figure()

        if all:
            n_plots = len(self.keys())
        else:
            n_plots = 1

        axes = fig.subplots(n_plots, sharex=True)
        if not isinstance(axes, Iterable):
            axes = [axes]

        keys_to_plot = self.keys() if all else [self.main_value_name]

        for ax, key in zip(axes, keys_to_plot):
            ax.plot(self.iters, self[key], label=key, *args, **kwargs)
            ax.set_ylabel(key)

            if xscale is not None:
                ax.set_xscale(xscale)
            if yscale is not None:
                if yscale == "auto":
                    if (
                        np.all(self[key] > 0)
                        and (np.max(self[key]) / np.min(self[key])) > 150
                    ):
                        yscale = "log"
                    else:
                        yscale = "linear"
                ax.set_yscale(yscale)

        plt.xlabel("Iterations")
        plt.legend()
        if show:
            plt.show()

        if n_plots == 1:
            return axes[0]
        else:
            return axes


@dispatch
def append(self: History, val: Any):
    return append(self, val, None)


@dispatch
def append(self: History, val: History, it: Any):  # noqa: E0102, F811
    if not set(self.keys()) == set(val.keys()):
        raise ValueError("cannot concatenate MVHistories with different keys")

    if it is not None:
        raise ValueError("When concatenating histories, cannot specify the iteration.")

    for key in self.keys():
        self._value_dict[key] = np.concatenate([self[key], val[key]])

    self._value_dict["iters"] = np.concatenate([self.iters, val.iters])
    return self


@dispatch
def append(self: History, values: dict, it: Any):  # noqa: E0102, F811
    for key, val in values.items():
        _vals = self._value_dict[key]

        if isinstance(_vals, list):
            _vals.append(val)
        elif isinstance(_vals, np.ndarray):
            new_shape = (len(_vals) + 1,) + _vals.shape[1:]
            # try to resize in place the buffer so that we don't reallocate
            # and if we fail, resize tby reallocating to a new buffer.
            try:
                _vals.resize(new_shape)
            except ValueError:
                _vals = np.resize(_vals, new_shape)
                self._value_dict[key] = _vals

            _vals[-1] = val
        else:
            raise TypeError(f"Unknown accumulator type {type(_vals)} for key {key}.")

    try:
        self.iters.resize(len(self.iters) + 1)
    except ValueError:
        self._value_dict["iters"] = np.resize(self.iters, (len(self.iters) + 1))

    self.iters[-1] = it
    return self


@dispatch
def append(self: History, val: Any, it: Any):  # noqa: E0102, F811
    if self._single_value and is_scalar(val) or hasattr(val, "__array__"):
        append(self, {"value": val}, it)
    elif hasattr(val, "to_compound"):
        append(self, val.to_compound()[1], it)
    elif hasattr(val, "to_dict"):
        append(self, val.to_dict(), it)
    else:
        append(self, {"value": val}, it)
    return self



# Copyright 2021 The NetKet Authors - All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING
from collections.abc import Callable

from netket.utils.struct import Pytree, field
from netket.utils.types import PyTree

from ._integrator_params import IntegratorParameters

if TYPE_CHECKING:
    from ._integrator_state import IntegratorState


class AbstractSolverState(Pytree):
    """
    Base class holding the state of a solver.
    """

    def __repr__(self):
        return "SolverState()"


class AbstractSolver(Pytree):
    r"""
    Abstract base class for ODE solvers.
    This object is an immutable pyTree. The structure used to hold any solver-specific data should be initialized by
    ``_init_state``.
    """

    dt: float = field(pytree_node=False)
    """The intial time-step size."""
    adaptive: bool = field(pytree_node=False, default=False)
    """The flag whether to use adaptive time-stepping."""
    integrator_params: IntegratorParameters = field(pytree_node=False)
    """Any additional arguments to pass to the Integrator"""

    def __init__(self, dt, adaptive=False, **kwargs):
        r"""
        Args:
            dt: The initial time-step size of the integrator.
            adaptive: A boolean indicator whether to use an adaptive scheme.

            atol: The tolerance for the absolute error on the solution if :code:`adaptive`.
                defaults to :code:`0.0`.
            rtol: The tolerance for the relative error on the solution if :code:`adaptive`.
                defaults to :code:`1e-7`.
            dt_limits: The extremal accepted values for the time-step size `dt` if :code:`adaptive`.
                defaults to :code:`(None, 10 * dt)`.
        """
        self.dt = dt
        self.adaptive = adaptive
        self.integrator_params = IntegratorParameters(dt=dt, **kwargs)

    def _init_state(self, integrator_state: "IntegratorState") -> AbstractSolverState:
        r"""
        Initializes the `SolverState` structure containing supplementary information needed.
        Args:
            integrator_state: The state of the Integrator

        Returns:
            An intialized `SolverState` instance
        """
        return AbstractSolverState()

    def step(
        self, f: Callable, dt: float, t: float, y_t: PyTree, state: AbstractSolverState
    ) -> tuple[PyTree, AbstractSolverState]:
        r"""
        Performs one fixed-size step from `t` to `t + dt`
        Args:
            f: The ODE function.
            dt: The current time-step size.
            t: The current time.
            y_t: The current solution.
            state: The state of the solver.

        Returns:
            The next solution y_t+1 and the corresponding updated state of the solver
        """

        raise NotImplementedError(
            "You need to define the method `step` in your `AbstractSolver`."
        )

    def step_with_error(
        self, f: Callable, dt: float, t: float, y_t: PyTree, state: AbstractSolverState
    ) -> tuple[PyTree, PyTree, AbstractSolverState]:
        r"""
        Perform one fixed-size step from `t` to `t + dt` and additionally returns the
        error vector provided by the adaptive solver.
        Args:
            f: The ODE function.
            dt: The current time-step size.
            t: The current time.
            y_t: The current solution.
            state: The state of the solver.

        Returns:
            The next solution y_t+1, the error y_err and the corresponding updated state of the solver
        """
        raise NotImplementedError(
            "You need to define the method `step_with_error` in your `AbstractSolver`."
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(dt={self.dt}, adaptive={self.adaptive}, integrator_parameters={self.integrator_params})"

    @property
    def is_explicit(self) -> bool:
        """Boolean indication whether the integrator is explicit."""
        raise NotImplementedError

    @property
    def is_adaptive(self) -> bool:
        """Boolean indication whether the integrator can be adaptive."""
        raise NotImplementedError

    @property
    def stages(self) -> int:
        """
        Number of stages (equal to the number of evaluations of the ode function) of the scheme.
        """
        raise NotImplementedError

    @property
    def is_fsal(self):
        """Returns True if the first iteration is the same as last."""
        # TODO: this is not yet supported
        return False


def append_docstring(doc):
    """
    Decorator that appends the string `doc` to the decorated function.

    This is needed here because docstrings cannot be f-strings or manipulated strings.
    """

    def _append_docstring(fun):
        fun.__doc__ = fun.__doc__ + doc
        return fun

    return _append_docstring


args_fixed_dt_docstring = """
    Args:
        dt: Timestep (floating-point number).
"""

args_adaptive_docstring = """
    This solver is adaptive, meaning that the time-step is changed at every
    iteration in order to keep the error below a certain threshold.

    In particular, given the variables at step :math:`t`, :math:`\\theta^{t}` and the
    error at the same time-step, :math:`\\epsilon^t`, we compute a rescaled error by
    using the absolute (**atol**) and relative (**reltol**) tolerances according
    to this formula.

    .. math::

        \\epsilon^\\text{scaled} = \\text{Norm}(\\frac{\\epsilon^{t}}{\\epsilon_{atol} +
            \\max(\\theta^t, \\theta^{t-1})\\epsilon_{reltol}}),

    where :math:`\\text{Norm}` is a function that normalises the vector, usually a vector
    norm but could be something else as well, and :math:`\\max` is an elementwise maximum
    function (with lexicographical ordering for complex numbers).

    Then, the integrator will attempt to keep `\\epsilon^\\text{scaled}<1`.

    Args:
        dt: Timestep (floating-point number). When :code:`adaptive==False` this value
            is never changed, when :code:`adaptive == True` this is the initial timestep.
        adaptive: Whether to use adaptive timestepping (Defaults to False).
            Not all integrators support adaptive timestepping.
        atol: Maximum absolute error at every time-step during adaptive timestepping.
            A larger value will lead to larger timestep. This option is ignored if
            `adaptive=False`. A value of 0 means it is ignored. Note that the `norm` used
            to compute the error can be  changed in the :ref:`netket.experimental.TDVP`
            driver. (Defaults to 0).
        rtol: Maximum relative error at every time-step during adaptive timestepping.
            A larger value will lead to larger timestep. This option is ignored if
            `adaptive=False`. Note that the `norm` used to compute the error can be
            changed in the :ref:`netket.experimental.TDVP` driver. (Defaults to 1e-7)
        dt_limits: A length-2 tuple of minimum and maximum timesteps considered by
            adaptive time-stepping. A value of None signals that there is no bound.
            Defaults to :code:`(None, 10*dt)`.
    """



# Copyright 2021-2023 The NetKet Authors - All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import jax.numpy as jnp

from numba import jit
from scipy.sparse import csr_matrix as _csr_matrix
from scipy.sparse import issparse

from netket import config
from netket.hilbert import DiscreteHilbert
from netket.operator import AbstractOperator
from netket.utils.optional_deps import import_optional_dependency


class DiscreteOperator(AbstractOperator[DiscreteHilbert]):
    r"""This class is the base class for operators defined on a
    discrete Hilbert space. Users interested in implementing new
    quantum Operators for discrete Hilbert spaces should derive
    their own class from this class
    """

    def __init__(self, hilbert: DiscreteHilbert):
        if not isinstance(hilbert, DiscreteHilbert):
            raise ValueError(
                "A Discrete Operator can only act upon a discrete Hilbert space."
            )
        super().__init__(hilbert)

    @property
    def max_conn_size(self) -> int:
        """The maximum number of non zero ⟨x|O|x'⟩ for every x."""
        raise NotImplementedError

    def get_conn_padded(self, x: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        r"""Finds the connected elements of the Operator.

        Starting from a batch of quantum numbers :math:`x={x_1, ... x_n}` of
        size :math:`B \times M` where :math:`B` size of the batch and :math:`M`
        size of the hilbert space, finds all states :math:`y_i^1, ..., y_i^K`
        connected to every :math:`x_i`.

        Returns a matrix of size :math:`B \times K_{max} \times M` where
        :math:`K_{max}` is the maximum number of connections for every
        :math:`y_i`.

        Args:
            x : A N-tensor of shape :math:`(...,hilbert.size)` containing
                the batch/batches of quantum numbers :math:`x`.

        Returns:
            **(x_primes, mels)**: The connected states x', in a N+1-tensor and an
            N-tensor containing the matrix elements :math:`O(x,x')`
            associated to each x' for every batch.
        """
        if config.netket_experimental_sharding:
            raise RuntimeError(
                "When using Sharding mode, only jax operators are supported."
            )

        n_visible = x.shape[-1]
        n_samples = x.size // n_visible

        sections = np.empty(n_samples, dtype=np.int32)
        x_primes, mels = self.get_conn_flattened(
            x.reshape(-1, x.shape[-1]), sections, pad=True
        )

        n_primes = sections[0]

        x_primes_r = x_primes.reshape(*x.shape[:-1], n_primes, n_visible)
        mels_r = mels.reshape(*x.shape[:-1], n_primes)

        return x_primes_r, mels_r

    def get_conn_flattened(
        self, x: np.ndarray, sections: np.ndarray, pad: bool = False
    ) -> tuple[np.ndarray, np.ndarray]:
        r"""Finds the connected elements of the Operator.

        Starting from a given quantum number :math:`x`, it finds all
        other quantum numbers  :math:`x'` such that the matrix element
        :math:`O(x,x')` is different from zero. In general there will be
        several different connected states :math:`x'` satisfying this
        condition, and they are denoted here :math:`x'(k)`, for
        :math:`k=0,1...N_{\mathrm{connected}}`.

        This is a batched version, where x is a matrix of shape
        :code:`(batch_size,hilbert.size)`.

        Args:
            x: A matrix of shape `(batch_size, hilbert.size)`
                containing the batch of quantum numbers x.
            sections: An array of sections for the flattened x'.
                See numpy.split for the meaning of sections.
            pad: If True, the output x' is padded to the maximum
                number of connected states. If False, the output x'
                is not padded and the number of connected states
                is given by the sections array. (Defaults: ``False``)

        Returns:
            (matrix, array): The connected states x', flattened together in
                a single matrix.
                An array containing the matrix elements :math:`O(x,x')`
                associated to each x'.

        """
        raise NotImplementedError(
            f"""
            The method get_conn_flattened has not been implemented for the object of
            type {type(self)}.

            This may happen if you defined a custom class inheriting from DiscreteOperator
            and you have not implemented this method. In that case, you should define
            `get_conn_flattened(self, x: array, sections: array)` according to the
            docstring provided on the documentation.

            Otherwise, please open an issue on netket's github repository.
            """
        )

    def get_conn(self, x: np.ndarray):
        r"""Finds the connected elements of the Operator. Starting
        from a given quantum number x, it finds all other quantum numbers x' such
        that the matrix element :math:`O(x,x')` is different from zero. In general there
        will be several different connected states x' satisfying this
        condition, and they are denoted here :math:`x'(k)`, for :math:`k=0,1...N_{\mathrm{connected}}`.

        Args:
            x: An array of shape `(hilbert.size, )` containing the quantum numbers x.

        Returns:
            matrix: The connected states x' of shape (N_connected,hilbert.size)
            array: An array containing the matrix elements :math:`O(x,x')` associated to each x'.

        Raise:
            ValueError: If the given quantum number is not compatible with the hilbert space.
        """
        if x.ndim != 1:
            raise ValueError(
                "get_conn does not support batches. Please use get_conn_flattened instead."
            )
        if x.shape[0] != self.hilbert.size:
            raise ValueError(
                "The given quantum numbers do not match the hilbert space because"
                f"it has shape {x.shape} of which[0] but expected {self.hilbert.size}."
            )

        return self.get_conn_flattened(
            x.reshape((1, -1)),
            np.ones(1),
        )

    def n_conn(self, x, out=None) -> np.ndarray:
        r"""Return the number of states connected to x.

        Args:
            x (matrix): A matrix of shape (batch_size,hilbert.size) containing
                        the batch of quantum numbers x.
            out (array): If None an output array is allocated.

        Returns:
            array: The number of connected states x' for each x[i].

        """
        if out is None:
            out = np.empty(x.shape[0], dtype=np.int32)
        self.get_conn_flattened(x, out)
        out = self._n_conn_from_sections(out)

        return out

    @staticmethod
    @jit(nopython=True)
    def _n_conn_from_sections(out):
        low = 0
        for i in range(out.shape[0]):
            old_out = out[i]
            out[i] = out[i] - low
            low = old_out

        return out

    def to_sparse(self) -> _csr_matrix:
        r"""Returns the sparse matrix representation of the operator. Note that,
        in general, the size of the matrix is exponential in the number of quantum
        numbers, and this operation should thus only be performed for
        low-dimensional Hilbert spaces or sufficiently sparse operators.

        This method requires an indexable Hilbert space.

        Returns:
            The sparse matrix representation of the operator.
        """
        concrete_op = self.collect()
        hilb = self.hilbert

        x = hilb.all_states()

        sections = np.empty(x.shape[0], dtype=np.int32)
        x_prime, mels = concrete_op.get_conn_flattened(x, sections)

        numbers = hilb.states_to_numbers(x_prime)

        sections1 = np.empty(sections.size + 1, dtype=np.int32)
        sections1[1:] = sections
        sections1[0] = 0

        ## eliminate duplicates from numbers
        # rows_indices = compute_row_indices(hilb.states_to_numbers(x), sections1)

        return _csr_matrix(
            (mels, numbers, sections1),
            shape=(self.hilbert.n_states, self.hilbert.n_states),
        )

        # return _csr_matrix(
        #    (mels, (rows_indices, numbers)),
        #    shape=(self.hilbert.n_states, self.hilbert.n_states),
        # )

    def to_dense(self) -> np.ndarray:
        r"""Returns the dense matrix representation of the operator. Note that,
        in general, the size of the matrix is exponential in the number of quantum
        numbers, and this operation should thus only be performed for
        low-dimensional Hilbert spaces or sufficiently sparse operators.

        This method requires an indexable Hilbert space.

        Returns:
            The dense matrix representation of the operator as a Numpy array.
        """
        return self.to_sparse().todense().A

    def to_qobj(self):  # -> "qutip.Qobj"
        r"""Convert the operator to a qutip's Qobj.

        Returns:
            A :class:`qutip.Qobj` object.
        """
        qutip = import_optional_dependency("qutip", descr="to_qobj")

        return qutip.Qobj(
            self.to_sparse(), dims=[list(self.hilbert.shape), list(self.hilbert.shape)]
        )

    def __call__(self, v: np.ndarray) -> np.ndarray:
        return self.apply(v)

    def apply(self, v: np.ndarray) -> np.ndarray:
        op = self.to_linear_operator()
        return op @ v

    def __matmul__(self, other):
        if (
            isinstance(other, np.ndarray)
            or isinstance(other, jnp.ndarray)
            or issparse(other)
        ):
            return self.apply(other)
        elif isinstance(other, AbstractOperator):
            return self._op__matmul__(other)
        else:
            return NotImplemented

    def _op__matmul__(self, other):
        "Implementation on subclasses of __matmul__"
        return NotImplemented

    def __rmatmul__(self, other):
        if (
            isinstance(other, np.ndarray)
            or isinstance(other, jnp.ndarray)
            or issparse(other)
        ):
            # return self.apply(other)
            return NotImplemented
        elif isinstance(other, AbstractOperator):
            return self._op__rmatmul__(other)
        else:
            return NotImplemented

    def _op__rmatmul__(self, other):
        "Implementation on subclasses of __matmul__"
        return NotImplemented

    def to_linear_operator(self):
        return self.to_sparse()
